from __future__ import print_function

import boto3
import boto3_session_cache
import botocore.session

from . import config


_sessions = {}


def session():
    aws_config = config.get('aws') or {}
    profile = aws_config.pop('profile_name', None)
    if 'botocore' not in _sessions:
        print('[ssha] creating aws session')
        _sessions['botocore'] = botocore.session.Session(profile=profile)
        resolver = _sessions['botocore'].get_component('credential_provider')
        provider = resolver.get_provider('assume-role')
        provider.cache = boto3_session_cache.JSONFileCache()
    return boto3.Session(botocore_session=_sessions['botocore'], **aws_config)


def ec2():
    return session().client('ec2')


def ssm():
    return session().client('ssm')



import copy
import re
import subprocess

from . import errors, settings


_config = {}


def _exec(command):
    return subprocess.check_output(command, shell=True).strip()


def _exec_all(data):
    if isinstance(data, basestring):
        for var in re.findall('\$\{(.+?)\}', data):
            data = data.replace('${' + var + '}', get(var))
        if data.startswith('$(') and data.endswith(')'):
            data = _exec(data[2:-1])
    elif isinstance(data, dict):
        for key, value in data.items():
            data[key] = _exec_all(value)
    elif isinstance(data, list):
        data = [_exec_all(item) for item in data]
    return data


def _merge(target, source):
    for key in source:
        if key in target:
            if isinstance(target[key], dict) and isinstance(source[key], dict):
                _merge(target[key], source[key])
            else:
                target[key] = source[key]
        else:
            target[key] = source[key]


def get(key):

    keys = key.split('.')
    value = _config
    for key in keys:
        value = value.get(key)
        if value is None:
            return None

    if isinstance(value, (dict, list)):
        value = copy.deepcopy(value)

    return _exec_all(value)


def load(name):
    """
    Loads a config from the settings.

    """

    _config.update(settings.all())
    config_specific_settings = _config.pop('config', None) or {}

    if name:
        if name not in names():
            errors.string_exit('config {} not found in .ssha file'.format(name))
        if name in config_specific_settings:
            _merge(_config, config_specific_settings[name])
        _merge(_config, {'config': {'name': name}})


def names():
    ssha_settings = settings.all().get('ssha') or {}
    return ssha_settings.get('configs') or []



from __future__ import print_function

import os

from . import config


def _get_address(instance):
    hostname = instance.get('PublicIpAddress') or instance['PrivateIpAddress']
    username = config.get('ssh.username')
    if username:
        return username + '@' + hostname
    else:
        return hostname


def connect(instance, bastion):

    command = ['ssh']

    if config.get('verbose'):
        command += ['-v']

    identity_file = config.get('ssh.identity_file')
    if identity_file:
        command += ['-i', identity_file]

    if bastion:
        command += ['-A', '-t', _get_address(bastion), 'ssh']

    command += [_get_address(instance)]

    print('[ssha] running {}'.format(' '.join(command)))

    os.execlp('ssh', *command)



import collections
import curses

from curses import panel

from . import config, ec2


Item = collections.namedtuple('Item', field_names=('label', 'value'))


class Menu(object):

    def __init__(self, title, items, stdscreen):
        self.window = stdscreen.subwin(0, 0)
        self.window.timeout(1000)
        self.window.keypad(1)
        self.panel = panel.new_panel(self.window)
        self.panel.hide()
        panel.update_panels()

        self.title = title

        self.position = 0
        self.items = items

    def addstr(self, y, x, str, attr=None):
        if y + 1 < self.max_y and x + len(str) < self.max_x:
            self.window.addstr(y, x, str, attr)

    def navigate(self, n):
        self.position += n
        if self.position < 0:
            self.position = len(self.items) - 1
        elif self.position >= len(self.items):
            self.position = 0

    def display(self):

        self.panel.top()
        self.panel.show()
        self.window.clear()

        while True:

            self.window.refresh()
            curses.doupdate()

            self.max_y, self.max_x = self.window.getmaxyx()

            # Display the menu title.
            if self.title:
                self.addstr(1, 2, self.title, curses.A_NORMAL)
                self.addstr(2, 2, '-' * len(self.title), curses.A_NORMAL)
                offset_y = 3
            else:
                offset_y = 1

            for index, item in enumerate(self.items):

                # Highlight the selected item.
                if index == self.position:
                    mode = curses.A_REVERSE
                else:
                    mode = curses.A_NORMAL

                # Display the item.
                self.addstr(
                    offset_y + index,
                    2,
                    item.label,
                    mode,
                )

                # Write out spaces to handle when a message becomes shorter.
                self.addstr(
                    offset_y + index,
                    2 + len(item.label),
                    ' ' * 100,
                    curses.A_NORMAL,
                )

            # Because window.timeout was called,
            # this returns -1 if nothing was pressed.
            key = self.window.getch()

            if key in [curses.KEY_ENTER, ord('\n')]:
                return self.items[self.position].value
            elif key == curses.KEY_UP:
                self.navigate(-1)
            elif key == curses.KEY_DOWN:
                self.navigate(1)
            elif key in (81, 113):
                # Either q or Q was pressed.
                raise KeyboardInterrupt

        self.window.clear()
        self.panel.hide()
        panel.update_panels()
        curses.doupdate()


def _display(screen, items, title=None):
    curses.curs_set(0)
    menu = Menu(title, items, screen)
    return menu.display()


def choose_config(names, search):
    if search:
        return search
    elif len(names) > 1:
        items = [Item(label=name, value=name) for name in names]
        return curses.wrapper(_display, items)
    elif names:
        return names[0]
    else:
        return None


def choose_instance(instances, search):

    items = [Item(label=ec2.label(inst), value=inst) for inst in instances]

    if search:
        search = search.lower()
        items = [item for item in items if search in item.label.lower()]
        if len(items) == 1:
            return items[0].value

    if not items:
        return None

    return curses.wrapper(_display, items)



from __future__ import print_function

import copy
import os
import toml

from . import errors


_settings = {}


def _find_settings_path():

    cwd = os.path.realpath(os.getcwd())

    while cwd and cwd != '/':
        path = os.path.join(cwd, '.ssha')
        if os.path.isfile(path):
            return path
        cwd = os.path.dirname(cwd)

    return None


def _load(path):
    try:
        with open(path) as settings_file:
            return toml.loads(settings_file.read())
    except IOError as error:
        errors.string_exit('Error reading settings: {}'.format(error))
    except Exception as error:
        errors.string_exit('Error parsing settings: {}'.format(error))


def all():
    return copy.deepcopy(_settings)


def load(**defaults):
    if defaults.get('verbose'):
        print('[ssha] finding settings file')
    _settings.update(defaults)
    path = _find_settings_path()
    if path:
        if defaults.get('verbose'):
            print('[ssha] loading {}'.format(path))
        data = _load(path)
        _settings.update(data)
    else:
        errors.string_exit('Could not find .ssha file in current directory or parent directories')



from __future__ import print_function

from . import aws, config, errors, ssm


_instances = []


def _describe_instances():

    ec2 = aws.ec2()

    print('[ssha] discovering ec2 instances')

    next_token = None
    while True:

        kwargs = {}
        if next_token:
            kwargs['NextToken'] = next_token

        response = ec2.describe_instances()
        if response['ResponseMetadata']['HTTPStatusCode'] != 200:
            errors.json_exit(response)

        for reservation in response['Reservations']:
            for instance in reservation['Instances']:
                yield instance

        next_token = response.get('NextToken')
        if not next_token:
            break


def _filter_instances(instances, rules):
    result = []
    for instance in instances:
        if _rules_pass(instance, rules):
            result.append(instance)
    return result


def _instance_sort_key(instance):
    result = []
    for field in config.get('display.sort') or ['InstanceId']:
        value = instance
        for key in field.split('.'):
            value = value.get(key)
            if not value:
                break
        result.append(value or '')
    return result


def _rules_pass(obj, rules):

    for key, expected_value in rules.items():

        if key not in obj:
            return False

        if isinstance(expected_value, dict):
            nested_rules = expected_value
            if not _rules_pass(obj[key], nested_rules):
                return False

        elif obj[key] != expected_value:
            return False

    return True


def _find_instances(ec2_filters, ssm_filters):

    if not _instances:
        for instance in _describe_instances():
            tags = {}
            for tag in instance.get('Tags') or []:
                tags[tag['Key']] = tag['Value']
            instance['Tags'] = tags
            _instances.append(instance)

    instances = _instances

    if instances and ec2_filters:
        instances = _filter_instances(instances, ec2_filters)

    if instances and ssm_filters:
        for info in ssm.find_instances():
            for instance in instances:
                if instance['InstanceId'] == info['InstanceId']:
                    instance.update(info)
                    break
        instances = _filter_instances(instances, ssm_filters)

    return instances


def discover_bastion(instance):
    bastions = _find_instances(
        ec2_filters=config.get('bastion.ec2'),
        ssm_filters=config.get('bastion.ssm'),
    )
    for bastion in bastions:
        if bastion['InstanceId'] == instance['InstanceId']:
            return None
    if bastions:
        return bastions[0]
    else:
        errors.string_exit('Bastion not found')


def discover_instances():
    instances = _find_instances(
        ec2_filters=config.get('discover.ec2'),
        ssm_filters=config.get('discover.ssm'),
    )
    return sorted(instances, key=_instance_sort_key)


def label(instance):
    result = []
    for field in config.get('display.fields') or []:
        value = instance
        for key in field.split('.'):
            value = value.get(key)
        if value:
            result.append(value)
    return ' '.join(result) or instance['InstanceId']



from __future__ import print_function

import argparse

from . import config, ec2, menu, settings, ssm, ssh


try:

    parser = argparse.ArgumentParser(prog=__package__)
    parser.add_argument('config', nargs='?', help='Configuration name')
    parser.add_argument('search', nargs='?', help='Instance search string')
    parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode')

    args = parser.parse_args()

    settings.load(verbose=args.verbose)

    configs = config.names()
    config_name = menu.choose_config(configs, args.config)
    config.load(config_name)

    instances = ec2.discover_instances()
    instance = menu.choose_instance(instances, args.search)
    if instance:

        bastion = ec2.discover_bastion(instance)

        if config.get('ssm'):
            ssm.send_command(instance, bastion)

        ssh.connect(instance, bastion)

    else:
        print('[ssha] no matching instances found')

except KeyboardInterrupt:
    pass



#!/usr/bin/env python

from setuptools import setup

setup(
    name='ssha',
    version='0.4.1',
    description='SSH into AWS EC2 instances',
    author='Raymond Butcher',
    author_email='ray.butcher@claranet.uk',
    url='https://github.com/claranet/ssha',
    license='MIT License',
    packages=(
        'ssha',
    ),
    scripts=(
        'bin/ssha',
    ),
    install_requires=(
        'boto3',
        'boto3-session-cache',
        'toml.py',
    ),
)
